
-- ORIGINAL HEADER (new added below)
--------------------------------------------------------------------
-- Company       : XESS Corp.
-- Engineer      : Dave Vanden Bout
-- Creation Date : 05/17/2005
-- Copyright     : 2005, XESS Corp
-- Tool Versions : WebPACK 6.3.03i
--
-- Description:
--    Interface that allows reading and writing data between the PC parallel port
--    and the RAM on the XSA Board.
--
-- Revision:
--    1.1.0
--
-- Additional Comments:
--    1.1.0:
--        Separated SDRAM controller module from RAM interface.
--    1.0.0:
--        Initial release.
--
-- License:
--    This code can be freely distributed and modified as long as
--    this header is not removed.
--------------------------------------------------------------------

--------------------------------------------------------------------
-- Company       : Tallinn University of Technology
-- Engineer      : Anti Sullin
-- Creation Date : 05/03/2006
-- Tool Versions : WebPACK 8.1i
--
-- Description:
--    Interface, that allows reading and writing SRAM from PicoBlaze microcontroller
--       Based on Parallel port interface by XESS Corp.
--
-- Revision:
--    1.0.0: Initial release
--------------------------------------------------------------------


library IEEE, unisim;
use IEEE.std_logic_1164.all;

package ramintfc_pckg is

  component ramintfc_core
    generic(
      DATA_WIDTH   :     natural := 16;  -- width of data
      HADDR_WIDTH  :     natural := 24  -- host-side address width
      );
    port(
      -- picoblaze
		cpuclk		  : in std_logic;
		port_id       : in std_logic_vector(2 downto 0); -- port address
		out_port      : in std_logic_vector(7 downto 0); -- data out
		in_port       : inout std_logic_vector(7 downto 0); -- data in
		write_strobe  : in std_logic;		-- write strobe
		read_strobe   : in std_logic;		-- read strobe
		
		-- sdram
      clkin        : in  std_logic;     -- main clock input
      clk1x        : in  std_logic;     -- locked clock from SDRAM controller
      sdram_reset  : out std_logic;     -- reset to SDRAM controller
      lock         : in  std_logic;     -- indicates SDRAM clock DLLs have locked
      rd           : out std_logic;     -- read control signal to host-side of SDRAM controller
      wr           : out std_logic;     -- write control signal to host-side of SDRAM controller
      earlyOpBegun : in  std_logic;     -- indicates when read/write operation has begun
      rdDone       : in  std_logic;     -- indicates when SDRAM read/write operation is done
      hAddr        : out std_logic_vector(HADDR_WIDTH-1 downto 0);  -- host-side address to SDRAM controller
      hDIn         : out std_logic_vector(DATA_WIDTH-1 downto 0);  -- data to host-side of SDRAM controller
      hDOut        : in  std_logic_vector(DATA_WIDTH-1 downto 0)  -- data from host-side of SDRAM controller
      );
  end component;

end package ramintfc_pckg;



library IEEE, unisim;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use IEEE.numeric_std.all;
use unisim.vcomponents.all;
use WORK.common.all;

entity ramintfc_core is
  generic(
    DATA_WIDTH   :     natural := 16;   -- width of data
    HADDR_WIDTH  :     natural := 24    -- host-side address width
    );
  port(
    -- picoblaze
	 cpuclk	 	   : in std_logic;	 
	 port_id       : in std_logic_vector(2 downto 0); -- port address
	 out_port      : in std_logic_vector(7 downto 0); -- data out
	 in_port       : out std_logic_vector(7 downto 0); -- data in
	 write_strobe  : in std_logic;		-- write strobe
	 read_strobe   : in std_logic;		-- read strobe
	
  	 -- sdram
    clkin        : in  std_logic;       -- main clock input
    clk1x        : in  std_logic;       -- locked clock from SDRAM controller
    sdram_reset  : out std_logic;       -- reset to SDRAM controller
    lock         : in  std_logic;       -- indicates SDRAM clock DLLs have locked
    rd           : out std_logic;       -- read control signal to host-side of SDRAM controller
    wr           : out std_logic;       -- write control signal to host-side of SDRAM controller
    earlyOpBegun : in  std_logic;       -- indicates when read/write operation has begun
    rdDone       : in  std_logic;       -- indicates when SDRAM read operation is done
    hAddr        : out std_logic_vector(HADDR_WIDTH-1 downto 0);  -- host-side address to SDRAM controller
    hDIn         : out std_logic_vector(DATA_WIDTH-1 downto 0);  -- data to host-side of SDRAM controller
    hDOut        : in  std_logic_vector(DATA_WIDTH-1 downto 0)  -- data from host-side of SDRAM controller
    );
end entity;

architecture arch of ramintfc_core is

  signal data_in              : std_logic_vector(15 downto 0);
  
  signal rd_x						: std_logic; -- R/W signals to SDRAM controller
  signal wr_x						: std_logic;
  signal rd_l						: std_logic;
  signal wr_l						: std_logic;

  signal op_ack					: std_logic; -- R/W opBegun ack to picoblaze

  signal rdDone_x					: std_logic; -- rdDone cached copy for picoblaze 
  signal rdDone_ack				: std_logic; -- rdDone read ack from picoblaze
  
  signal sdram_reset_r			: std_logic; -- SDRAM reset signal
begin

  -- release the SDRAM controller reset once the SDRAM clock has stabilized
  process(clk1x)
  begin
    if(clk1x'event and clk1x = HI) then
      if lock = HI then
        sdram_reset_r <= LO;
      end if;
    end if;
  end process;
  sdram_reset         <= sdram_reset_r;


	-- Picoblaze i/o registers
  IO_registers: process(cpuclk)
  begin
    if cpuclk'event and cpuclk='1' then
					-- ADDRESS 7:0
		      if write_strobe='1' and port_id="000" then
					hAddr(7 downto 0) <= out_port;
					if rdDone_x='1' then
						rdDone_ack <='1';
					end if;
				end if;
					
					-- ADDRESS 15:8
		      if write_strobe='1' and port_id="001" then
					hAddr(15 downto 8) <= out_port;
					if rdDone_x='1' then
						rdDone_ack <='1';
					end if;
				end if;
				
					-- ADDRESS 23:16
		      if write_strobe='1' and port_id="010" then
					hAddr(23 downto 16) <= out_port;
					if rdDone_x='1' then
						rdDone_ack <='1';
					end if;

				end if;

					-- DATA 7:0
		      if write_strobe='1' and port_id="100" then
					hDin(7 downto 0) <= out_port;
				end if;
					
					-- DATA 15:8
		      if write_strobe='1' and port_id="101" then
					hDin(15 downto 8) <= out_port;
				end if;
					
					-- R/W signal async ack from ram
				if op_ack = '1' then
					rd_l <= '0';
					wr_l <= '0';
				end if;
				
					-- CONTROL PORT
		      if write_strobe='1' and port_id="111" then
					rd_l <= out_port(0);
					wr_l <= out_port(1);
				end if;

					-- rdDone async ack
				if rdDone_x='0' then
					rdDone_ack <='0';
				end if;
	end if;
end process;

-- PicoBlaze input mux
process(port_id, data_in, rdDone_x, rd_x, wr_x)
begin
				in_port <= (others=>'0');
				if port_id="100" then
					in_port <= data_in(7 downto 0);
				end if;
				if port_id="101" then
					in_port <= data_in(15 downto 8);
				end if;
				if port_id="111" then
					in_port <= "00000" & wr_x & rd_x & rdDone_x;
				end if;
end process;


  process(clk1x)
  begin
    if rising_edge(clk1x) then
		-- async r/w commands from picoblaze
		if earlyOpBegun='1' then
			rd_x<='0';
			wr_x<='0';
			op_ack<='1';
		else
			rd_x <= rd_l;
			wr_x <= wr_l;
		end if;
		if rd_l='0' and wr_l='0' then
			op_ack<='0';
		end if;
		
      -- fetch data when acknowledgement comes
		if rdDone = HI then
        data_in <= hDout;
		  rdDone_x <= '1';
      end if;
      if rdDone_ack = '1' then
			rdDone_x <= '0';
		end if;

    end if;
  end process;
  rd <= rd_x;
  wr <= wr_x;


end architecture;
