--------------------------------------------------------------------
-- Company       : Tallinn University of Technology
-- Engineer      : Anti Sullin
-- Creation Date : 05/03/2006
-- Tool Versions : WebPACK 8.1i
--
-- Description:
--		VIDEO GAME PLATFORM EXAMPLE. Includes snake demo.
--
-- Revision:
--    1.0.0: Initial release
--------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use IEEE.numeric_std.all;
use work.vga_pckg.all;
use work.xsasdram_ppintfc_pckg.all;
use work.common.all;


entity test_vga is
  generic(
    SDRAM_NROWS     :       natural := 8192;  -- 4096 for XSA-50, XSA-100; 8192 for XSA-200, XSA-3S1000
    SDRAM_NCOLS     :       natural := 512;  -- 256 for XSA-50; 512 for XSA-100, XSA-200, XSA-3S1000
    DATA_WIDTH      :       natural := 16;  -- SDRAM databus width
    SADDR_WIDTH     :       natural := 13;  -- # of SDRAM address bits
    HADDR_WIDTH     :       natural := 24;  -- host-side address width
    FREQ            :       natural := 100_000;  -- 50 MHz for XSA-50, XSA-100; 100 MHz for XSA-200, XSA-3S1000
    CLK_DIV         :       natural := 4;  -- pixel clock = FREQ / CLK_DIV
    PIXEL_WIDTH     :       natural := 8;  -- width of a pixel in memory
    NUM_RGB_BITS    :       natural := 3;  -- #bits in each R,G,B component of a pixel
    PIXELS_PER_LINE :       natural := 640;  -- width of image in pixels
    LINES_PER_FRAME :       natural := 480;  -- height of image in scanlines
    FIT_TO_SCREEN   :       boolean := true;  -- adapt video timing to fit image width x height
	 MULTIPLE_ACTIVE_ROWS :  boolean := true
    );
  port(
    rst_n           : in    std_logic;  -- reset

    clk             : in    std_logic;  -- master clock (frequency set by FREQ)
	
	 -- VGA OUT
    vsync_n         : out   std_logic;  -- VGA vertical sync
    hsync_n         : out   std_logic;  -- VGA horizontal sync
    red             : out   std_logic_vector(NUM_RGB_BITS-1 downto 0);  -- VGA red signals
    green           : out   std_logic_vector(NUM_RGB_BITS-1 downto 0);  -- VGA green signals
    blue            : out   std_logic_vector(NUM_RGB_BITS-1 downto 0);  -- VGA blue signals

	 -- PS2 KEYBOARD IN
	 ps2c				  : in	 std_logic; -- ps2 clock in
	 ps2d				  : in	 std_logic; -- ps2 clock in	 

    -- SDRAM I/O
    sclkfb          : in    std_logic;  -- clock from SDRAM after PCB delays
    sclk            : out   std_logic;  -- SDRAM clock sync'ed to master clock
    cke             : out   std_logic;  -- clock-enable to SDRAM
    cs_n            : out   std_logic;  -- chip-select to SDRAM
    ras_n           : out   std_logic;  -- SDRAM row address strobe
    cas_n           : out   std_logic;  -- SDRAM column address strobe
    we_n            : out   std_logic;  -- SDRAM write enable
    ba              : out   std_logic_vector(1 downto 0);  -- SDRAM bank address bits
    sAddr           : out   std_logic_vector(SADDR_WIDTH-1 downto 0);  -- SDRAM row/column address
    sData           : inout std_logic_vector(DATA_WIDTH-1 downto 0);  -- SDRAM in/out databus
    dqmh            : out   std_logic;  -- high databits I/O mask
    dqml            : out   std_logic   -- low databits I/O mask
    );
end entity;



architecture arch of test_vga is
  
  -- PICOBLAZE
  component kcpsm3 
    Port (      address : out std_logic_vector(9 downto 0);
            instruction : in std_logic_vector(17 downto 0);
                port_id : out std_logic_vector(7 downto 0);
           write_strobe : out std_logic;
               out_port : out std_logic_vector(7 downto 0);
            read_strobe : out std_logic;
                in_port : in std_logic_vector(7 downto 0);
              interrupt : in std_logic;
          interrupt_ack : out std_logic;
                  reset : in std_logic;
                    clk : in std_logic);
    end component;

  -- GAME ROM FOR PICOBLAZE
  component game 
    Port (      address : in std_logic_vector(9 downto 0);
            instruction : out std_logic_vector(17 downto 0);
                    clk : in std_logic);
    end component;

  -- KEYBOARD CONTROLLER
  component keyb
	Port (	ps2c		: in std_logic;
				ps2d		: in std_logic;
				clk		: in std_logic;
				btns		: out std_logic_vector(3 downto 0)); -- up, dn, left, right
  end component;				


  signal clk1x        : std_logic;      -- sync'ed clock from SDRAM controller
  signal cpuclk       : std_logic;      -- cpu clk
  signal rst          : std_logic;      -- reset signal
  signal eof          : std_logic;      -- end-of-frame signal from VGA controller
  signal earlyOpBegun : std_logic;      -- indicates when an SDRAM read operation has begun
  signal rdDone       : std_logic;      -- indicates when data read from the SDRAM is available
  signal full, full_n : std_logic;      -- indicates when the VGA pixel buffer is full
  signal address      : std_logic_vector(HADDR_WIDTH-1 downto 0);  -- SDRAM address value
  signal pixel        : std_logic_vector(DATA_WIDTH-1 downto 0);  -- pixel values from SDRAM


  signal clk2x			 : std_logic;      -- video aadressi valik
  signal adr         : std_logic_vector(HADDR_WIDTH-1 downto 0);  -- SDRAM address counter 

  -- PICOBLAZE connections
  signal romaddress       : std_logic_vector(9 downto 0);
  signal instruction   : std_logic_vector(17 downto 0);
  signal port_id       : std_logic_vector(7 downto 0);
  signal out_port      : std_logic_vector(7 downto 0);
  signal in_port0       : std_logic_vector(7 downto 0);
  signal in_port1       : std_logic_vector(7 downto 0);
  signal in_port       : std_logic_vector(7 downto 0);
  signal write_strobe  : std_logic;
  signal read_strobe   : std_logic;
  signal interrupt     : std_logic :='0';
  signal interrupt_ack : std_logic;
  signal eof0          : std_logic;      -- end-of-frame signal with 1 clk1x delay
  signal reset         : std_logic;
  signal write_strobe0	: std_logic;
  signal read_strobe0	: std_logic;

  -- VGA GENERATOR PARAMETERS
  signal opaquecolor      : std_logic_vector(7 downto 0);
  signal alpha1a          : std_logic_vector(3 downto 0);
  signal alpha1b          : std_logic_vector(3 downto 0);
  signal alpha2a          : std_logic_vector(3 downto 0);

signal btns				: std_logic_vector(3 downto 0);
  
begin

cpuclk <= clk1x;
reset <= rst;

  processor: kcpsm3
    port map(      address => romaddress,
               instruction => instruction,
                   port_id => port_id,
              write_strobe => write_strobe,
                  out_port => out_port,
               read_strobe => read_strobe,
                   in_port => in_port,
                 interrupt => interrupt,
             interrupt_ack => interrupt_ack,
                     reset => reset,
                       clk => cpuclk);
 
  program: game
    port map(      address => romaddress,
               instruction => instruction,
                       clk => cpuclk);



  --Interrupt at the end of every video frame
  interrupt_control: process(cpuclk)
  begin

    if cpuclk'event and cpuclk='1' then
      if interrupt_ack='1' then
        interrupt <= '0';
       elsif eof='1' and eof0='0' then -- eof edge
        interrupt <= '1';
       else
        interrupt <= interrupt;
      end if;
		eof0<=eof;
    end if; 

  end process interrupt_control;




  -- picoblaze output ports 1xxx
  IO_registers: process(cpuclk)
  begin

    if cpuclk'event and cpuclk='1' then
      if write_strobe='1' and port_id(3 downto 0)="1000" then
        opaquecolor <= out_port;
		end if;
		if write_strobe='1' and port_id(3 downto 0)="1001" then
        alpha1a <= out_port(3 downto 0);
		end if;
		if write_strobe='1' and port_id(3 downto 0)="1010" then
        alpha1b <= out_port(3 downto 0);
		end if;
		if write_strobe='1' and port_id(3 downto 0)="1011" then
        alpha2a <= out_port(3 downto 0);
		end if;
	end if;	
  end process IO_registers;

  -- picoblaze input ports 1xxx
  in_port1 <= "0000" & btns;


  -- mux picoblaze input ports 0xxx, 1xxx
  in_port <= in_port0 when port_id(3)='0' else in_port1;

  -- sdram io address demux for io ports 0xxx
  write_strobe0 <= write_strobe when port_id(3)='0' else '0';
  read_strobe0 <= read_strobe when port_id(3)='0' else '0';



  -- KEYBOARD CONTROLLER
  kbd: keyb
	 port map(
		ps2c => ps2c,
		ps2d => ps2d,
		clk => cpuclk,
		btns => btns);

  -- reset inversion
  rst <= not rst_n;

  -- SDRAM address counter for VGA generator
  process(clk1x)
  begin
    if rising_edge(clk1x) then
      if eof = YES then
		  clk2x <='0';
		  adr <= std_logic_vector(TO_UNSIGNED(1079, address'length));  -- reset the address at the end of a video frame
      elsif earlyOpBegun = YES then
			  adr <= adr + 1;         -- go to the next address once the read of the current address has begun
			  address<=adr;
      end if;
    end if;
  end process;

  -- XSA SDRAM controller used to get pixel data from the external SDRAM
  u0 : XSASDRAMCntl_ppintfc
    generic map(
      FREQ         => FREQ,
      PIPE_EN      => true,             -- use pipelining for maximum speed
      MAX_NOP      => 1000000,          -- disable self-refresh since it takes too long to re-awaken the SDRAM with video timing
      DATA_WIDTH   => DATA_WIDTH,
      NROWS        => SDRAM_NROWS,
      NCOLS        => SDRAM_NCOLS,
      HADDR_WIDTH  => HADDR_WIDTH,
      SADDR_WIDTH  => SADDR_WIDTH,
		MULTIPLE_ACTIVE_ROWS =>MULTIPLE_ACTIVE_ROWS
      )
    port map(
		cpuclk		 => cpuclk,
		
		-- picoblaze side
		port_id 		 => port_id(2 downto 0),
		out_port     => out_port,
		in_port      => in_port0,
		write_strobe => write_strobe0,
		read_strobe  => read_strobe0,

      -- host side
      clk          => clk,              -- master clock
      clk1x        => clk1x,            -- master clock resync'ed to account for delays to external SDRAM
      rst          => rst,
      rd           => full_n,           -- initiate a read when the VGA pixel buffer is not full
      hAddr        => address,          -- the address to read from is stored in the address counter
      earlyOpBegun => earlyOpBegun,     -- indicate when the read operation has actually begun
      rdDone       => rdDone,           -- indicate when the data from the read operation is available
      hDOut        => pixel,            -- this is the pixel data that was read from the SDRAM
      wr           => '0',              -- no SDRAM writing is needed in this application
      hDIn         => std_logic_vector(TO_UNSIGNED(0, DATA_WIDTH)),  -- set the SDRAM write-data bus to zeroes
      -- SDRAM side
      sclkfb       => sclkfb,
      sclk         => sclk,
      cke          => cke,
      cs_n         => cs_n,
      ras_n        => ras_n,
      cas_n        => cas_n,
      we_n         => we_n,
      ba           => ba,
      sAddr        => sAddr,
      sData        => sData,
      dqmh         => dqmh,
      dqml         => dqml
      );

  -- VGA generator
  u1 : vga
    generic map (
      FREQ            => FREQ,
      CLK_DIV         => CLK_DIV,
      PIXEL_WIDTH     => PIXEL_WIDTH,
      PIXELS_PER_LINE => PIXELS_PER_LINE,
      LINES_PER_FRAME => LINES_PER_FRAME,
      NUM_RGB_BITS    => NUM_RGB_BITS,
      FIT_TO_SCREEN   => FIT_TO_SCREEN
      )
    port map (
      rst             => rst,
      clk             => clk1x,
      wr              => rdDone,			 -- FIFO read signal
      pixel_data_in   => pixel,
      
		-- 2-layer display control
		opaquecolor		 => opaquecolor,
		alpha1a		    => alpha1a,
		alpha1b		    => alpha1b,
		alpha2a		    => alpha2a,
		
		full            => full,          -- FIFO is full
      eof             => eof,           -- end-of-frame
      r               => red,           -- RGB out
      g               => green,
      b               => blue,
      hsync_n         => hsync_n,       -- horizontal sync out
      vsync_n         => vsync_n,       -- vertical sync out
      blank           => open
      );


	process (clk1x)
	begin
		if rising_edge(clk1x) then
			full_n <= not full;  -- VGA FIFO full signal inverted buffer
		end if;
	end process;

end arch;
