-- ORIGINAL HEADER:
--------------------------------------------------------------------
-- Company       : XESS Corp.
-- Engineer      : Dave Vanden Bout
-- Creation Date : 05/17/2005
-- Copyright     : 2005, XESS Corp
-- Tool Versions : WebPACK 6.3.03i
--
-- Description:
--    VGA generator that displays pixels on a VGA monitor.
--
-- Revision:
--    1.0.0
--
-- Additional Comments:
--    1.0.0:
--        Initial release.
--
-- License:
--    This code can be freely distributed and modified as long as
--    this header is not removed.
--------------------------------------------------------------------


--------------------------------------------------------------------
-- Company       : Tallinn University of Technology
-- Engineer      : Anti Sullin
-- Creation Date : 04/22/2006
-- Tool Versions : WebPACK 8.1i
--
-- Description:
--    VGA GENERATOR, that is improved by adding dual layer support,
--    layer alpha controls and second layer opaque color support.
--       Based on VGA interface by XESS Corp.
--
-- Principe of operation:
-- 	SDRAM layout:
-- 		every first SDRAM word is one layer, second word is second layer.
--       every word contains two pixels.
--    layer support:
--       if pixel_on_layer2 = opaque_color then
--				r = r_layer1 * alpha1b
--				g = g_layer1 * alpha1b
--				b = b_layer1 * alpha1b
--			else
--				r = r_layer1 * alpha1a + r_layer2 * alpha2a
--				g = g_layer1 * alpha1a + g_layer2 * alpha2a
--				b = b_layer1 * alpha1a + b_layer2 * alpha2a
--			end if;
--		NB! addition does not overflow but saturates!
--
--
-- Revision:
--    2.0.0: Initial release
--------------------------------------------------------------------



library IEEE;
use IEEE.std_logic_1164.all;

package vga_pckg is
  component vga
    generic (
      FREQ            :     natural := 50_000;  -- master clock frequency (in KHz)
      CLK_DIV         :     natural := 1;  -- FREQ / CLK_DIV = pixel clock
      PIXEL_WIDTH     :     natural := 8;  -- pixel width: 1, 2, 4, 8, or 16 bits
      PIXELS_PER_LINE :     natural := 800;  -- pixels per video scan line
      LINES_PER_FRAME :     natural := 600;  -- scan lines per video frame
      NUM_RGB_BITS    :     natural := 2;  -- width of R, G and B color output buses (2 or 3 are only valid values)
      FIT_TO_SCREEN   :     boolean := true  -- fit width x length to monitor screen
      );
    port (
      rst             : in  std_logic;  -- reset
      clk             : in  std_logic;  -- master clock
      wr              : in  std_logic;  -- write-enable for pixel buffer
      pixel_data_in   : in  std_logic_vector(15 downto 0);  -- input databus to pixel buffer
		opaquecolor		 : in std_logic_vector(7 downto 0); -- opaque color of second layer
		alpha1a			 : in std_logic_vector(3 downto 0); -- alpha of layer 1 if second layer is opaque
		alpha1b			 : in std_logic_vector(3 downto 0); -- alpha of layer 1 if second layer is not opaque
		alpha2a			 : in std_logic_vector(3 downto 0); -- alpha of layer 2
      full            : out std_logic;  -- pixel buffer full
      eof             : out std_logic;  -- end of vga frame
      r, g, b         : out std_logic_vector(NUM_RGB_BITS-1 downto 0);  -- R,G,B color output buses
      hsync_n         : out std_logic;  -- horizontal sync pulse
      vsync_n         : out std_logic;  -- vertical sync pulse
      blank           : out std_logic   -- blanking signal
		
      );
  end component vga;
end package vga_pckg;



library IEEE, unisim;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use IEEE.std_logic_unsigned.all;
use unisim.vcomponents.all;
use work.common.all;

entity vga is
  generic (
    FREQ            :     natural := 50_000;  -- master clock frequency (in KHz)
    CLK_DIV         :     natural := 1;  -- FREQ / CLK_DIV = pixel clock
    PIXEL_WIDTH     :     natural := 4;  -- pixel width: 1, 2, 4, 8, or 16 bits
    PIXELS_PER_LINE :     natural := 1024;  -- pixels per video scan line
    LINES_PER_FRAME :     natural := 512;  -- scan lines per video frame
    NUM_RGB_BITS    :     natural := 2;  -- width of R, G and B color output buses
    FIT_TO_SCREEN   :     boolean := true  -- fit width x length to monitor screen
    );
  port (
    rst             : in  std_logic;    -- reset
    clk             : in  std_logic;    -- master clock
    wr              : in  std_logic;    -- write-enable for pixel buffer
    pixel_data_in   : in  std_logic_vector(15 downto 0);  -- input databus to pixel buffer
 	 opaquecolor		: in std_logic_vector(7 downto 0); -- opaque color of second layer
	 alpha1a			 : in std_logic_vector(3 downto 0);
	 alpha1b			 : in std_logic_vector(3 downto 0);
	 alpha2a			 : in std_logic_vector(3 downto 0);
    full            : out std_logic;    -- pixel buffer full
    eof             : out std_logic;    -- end of vga frame
    r, g, b         : out std_logic_vector(NUM_RGB_BITS-1 downto 0);  -- R,G,B color output buses
    hsync_n         : out std_logic;    -- horizontal sync pulse
    vsync_n         : out std_logic;    -- vertical sync pulse
    blank           : out std_logic     -- blanking signal
    );
end entity vga;


architecture vga_arch of vga is
  constant NORM : natural := 1000;      -- normalization factor for us * KHz

  -- video timing parameters for FIT_TO_SCREEN mode
  constant HSYNC_START_F  : natural := (NORM * PIXELS_PER_LINE * CLK_DIV)/FREQ + 1;  -- start of horiz. sync pulse with a scanline (us)
  constant HSYNC_PERIOD_F : natural := HSYNC_START_F + 6;  -- horizontal scanline period (us)
  constant HSYNC_WIDTH_F  : natural := 4;  -- width of horiz. sync pulse (us)
  constant HSYNC_FREQ_F   : natural := NORM / HSYNC_PERIOD_F;  -- scanline frequency (KHz) 
  constant VSYNC_START_F  : natural := HSYNC_PERIOD_F * LINES_PER_FRAME + 340;  -- start of vert. sync pulse within a frame (us)
  constant VSYNC_PERIOD_F : natural := VSYNC_START_F + 1084;  -- video frame period (us)
  constant VSYNC_WIDTH_F  : natural := 64;  -- width of vert. sync pulse (us)

  -- video timing for 31 KHz horizontal, 60 Hz vertical screen refresh
  constant HSYNC_START  : natural := 26;  -- start of horiz. sync pulse with a scanline (us)
  constant HSYNC_PERIOD : natural := 32;  -- horizontal scanline period (us)
  constant HSYNC_WIDTH  : natural := 4;  -- width of horiz. sync pulse (us)
  constant HSYNC_FREQ   : natural := NORM / HSYNC_PERIOD;  -- scanline frequency (KHz) 
  constant VSYNC_START  : natural := 15_700;  -- start of vert. sync pulse within a frame (us)
  constant VSYNC_PERIOD : natural := 16_784;  -- video frame period (us)
  constant VSYNC_WIDTH  : natural := 64;  -- width of vert. sync pulse (us)

  signal   clk_div_cnt                :     natural range 0 to 255;
  signal   cke                        :     std_logic;
  signal   line_cnt, pixel_cnt        :     std_logic_vector(15 downto 0);  -- current video line and pixel within line
  signal   fifo_rst                   :     std_logic;
  signal   fifo_level                 :     std_logic_vector(7 downto 0);
  signal   eof_i, eof_x, eof_r        :     std_logic;
  signal   v_gate, cke_v_gate         :     std_logic;
  signal   h_blank, v_blank, visible  :     std_logic;
  constant PIX_PROC_DELAY             :     natural := 5;  -- time delay to read a pixel from the FIFO and colormap it
  signal   hsync_x, hsync_r           :     std_logic_vector(PIX_PROC_DELAY downto 1);
  signal   blank_x, blank_r           :     std_logic_vector(PIX_PROC_DELAY downto 1);
  signal   cke_rd, rd_x, rd_r         :     std_logic;
  signal   pixel                      :     std_logic_vector(8 downto 0); -- 3x3 bits
  signal   pixel_data_x, pixel_data_r :     std_logic_vector(15 downto 0);
  signal   pixel_data_x1, pixel_data_r1 :     std_logic_vector(15 downto 0);
  signal   pixel_data_x2, pixel_data_r2 :     std_logic_vector(15 downto 0);
  signal   pixel_data_temp_x, pixel_data_temp_r :     std_logic_vector(15 downto 0);
  signal   pixel_data_out             :     std_logic_vector(15 downto 0);
  signal   rgb_x, rgb_r               :     std_logic_vector(3*NUM_RGB_BITS-1 downto 0);
  
  signal alpha1 : std_logic_vector(3 downto 0);
  signal alpha2 : std_logic_vector(3 downto 0);
  
  	signal temp_r1 : std_logic_vector(2 downto 0);
	signal temp_g1 : std_logic_vector(2 downto 0);
	signal temp_b1 : std_logic_vector(2 downto 0);
	signal temp_r2 : std_logic_vector(2 downto 0);
	signal temp_g2 : std_logic_vector(2 downto 0);
	signal temp_b2 : std_logic_vector(2 downto 0);
	signal temp_r : std_logic_vector(3 downto 0);
	signal temp_g : std_logic_vector(3 downto 0);
	signal temp_b : std_logic_vector(3 downto 0);
	
	signal satur_r : std_logic;
	signal satur_g : std_logic;
	signal satur_b : std_logic;

	signal temp_in_b1 : std_logic_vector(2 downto 0);
	signal temp_in_b2 : std_logic_vector(2 downto 0);

  -- pixel pipeline
  component fifo_cc
    port (
      clk                             : in  std_logic;
      rst                             : in  std_logic;
      rd                              : in  std_logic;
      wr                              : in  std_logic;
      data_in                         : in  std_logic_vector(15 downto 0);
      data_out                        : out std_logic_vector(15 downto 0);
      full                            : out std_logic;
      empty                           : out std_logic;
      level                           : out std_logic_vector(7 downto 0)
      );
  end component fifo_cc;

  -- 3-bit multiplier
  component mul3
	 	port (
		x	:	in std_logic_vector(3 downto 0);
		y	:	in	std_logic_vector(2 downto 0);
		p : out std_logic_vector(2 downto 0)
		);
	end component mul3;
	   
  -- sync generator
  component sync
    generic (
      FREQ                            :     natural := 50_000;  -- master clock frequency (in KHz)
      PERIOD                          :     natural := 32;  -- period of sync pulse (in us)
      START                           :     natural := 26;  -- time sync pulse starts within the period (in us)
      WIDTH                           :     natural := 4;  -- width of sync pulse (in us)
      VISIBLE                         :     natural := 1024  -- number of visible pixels/line or lines/frame
      );
    port (
      rst                             : in  std_logic;  -- reset
      clk                             : in  std_logic;  -- master clock
      cke                             : in  std_logic;  -- clock-enable
      sync_n                          : out std_logic;  -- sync pulse
      gate                            : out std_logic;  -- single-clock pulse at start of sync pulse
      blank                           : out std_logic;  -- blanking signal
      cnt                             : out std_logic_vector(15 downto 0)  -- output the timing counter value
      );
  end component sync;

begin

  -- clock divider for reducing the pixel clock rate
  process(clk, rst)
  begin
    if rst = YES then
      clk_div_cnt   <= 0;
      cke           <= YES;
    elsif rising_edge(clk) then
      if clk_div_cnt = CLK_DIV-1 then
        clk_div_cnt <= 0;
        cke         <= YES;
      else
        clk_div_cnt <= clk_div_cnt + 1;
        cke         <= NO;
      end if;
    end if;
  end process;

  -- pixel data buffer
  cke_rd <= rd_x and cke;
  fifo : fifo_cc
    port map (
      clk      => clk,
      rd       => cke_rd,
      wr       => wr,
      data_in  => pixel_data_in,
      rst      => fifo_rst,
      data_out => pixel_data_out,
      full     => open,
      empty    => open,
      level    => fifo_level
      );
  full   <= YES when fifo_level(7 downto 3) = "11111" else NO;

  -- the clock enable for the vertical sync module is also combined with a gate signal
  -- that is generated at the end of every scanline by the horizontal sync module
  cke_v_gate <= cke and v_gate;

  -- generate the horizontal and vertical sync pulses for FIT_TO_SCREEN mode
  hsync       : sync
      generic map (
        FREQ    => FREQ / CLK_DIV,      -- master pixel-clock frequency
        PERIOD  => HSYNC_PERIOD_F,      -- scanline period (32 us)
        START   => HSYNC_START_F,       -- start of horizontal sync pulse in scan line
        WIDTH   => HSYNC_WIDTH_F,       -- width of horizontal sync pulse
        VISIBLE => PIXELS_PER_LINE
        )
      port map (
        rst     => rst,
        clk     => clk,                 -- master clock
        cke     => cke,                 -- a new pixel is output whenever the clock is enabled
        sync_n  => hsync_x(1),          -- send pulse through delay line
        gate    => v_gate,              -- send gate signal to increment vertical sync pulse generator once per scan line
        blank   => h_blank,             -- blanking signal within a scan line
        cnt     => pixel_cnt            -- current pixel within the scan line
        );

    vsync : sync
      generic map (
        FREQ    => HSYNC_FREQ_F,        -- scanline frequency
        PERIOD  => VSYNC_PERIOD_F,      -- image frame period
        START   => VSYNC_START_F,       -- start of vertical sync pulse in frame
        WIDTH   => VSYNC_WIDTH_F,       -- width of vertical sync pulse
        VISIBLE => LINES_PER_FRAME
        )
      port map (
        rst     => rst,
        clk     => clk,                 -- master clock
        cke     => cke_v_gate,          -- enable clock once per horizontal scan line
        sync_n  => vsync_n,             -- send pulse through delay line
        gate    => eof_x,               -- indicate the end of a complete frame
        blank   => v_blank,             -- blanking signal within a frame
        cnt     => line_cnt             -- current scan line within a frame
        );


  eof_i    <= eof_x and not eof_r;      -- shorten end-of-frame signal to a single clock cycle
  eof      <= eof_i;
  fifo_rst <= eof_i or rst;             -- clear the contents of the pixel buffer at the end of every frame

  visible    <= h_blank nor v_blank;    -- pixels are visible when horiz. & vertical blank are inactive 
  blank_x(1) <= not visible;            -- send blanking signal through delay line

  -- pass the horiz. and vert. syncs and blanking signal through delay lines to compensate for the
  -- processing delays incurred by the pixel data
  hsync_x(hsync_x'high downto 2) <= hsync_r(hsync_r'high-1 downto 1);
  hsync_n                        <= hsync_r(hsync_r'high);
  blank_x(blank_x'high downto 2) <= blank_r(blank_r'high-1 downto 1);
  blank                          <= blank_r(blank_r'high);






  -- get the current pixel from the word of pixel data or read more pixel data from the buffer
  get_pixel : process(visible, pixel_data_out, pixel_data_r, rd_r, pixel_cnt)

  begin
    rd_x <= NO;
	 if (visible = YES) then
      rd_x       <= YES;     -- read new pixel data only during visible portion of scan line
    end if;
         
	-- store read pixel data and reorder both layers into one word from consequtive SDRAM words
	 if rd_r = YES then
		if pixel_cnt(0 downto 0) = 0 then
			pixel_data_temp_x <= pixel_data_out;
			pixel_data_x <= pixel_data_r1(15 downto 8) & pixel_data_r2(15 downto 8);
		else
			pixel_data_x2 <= pixel_data_out;
			pixel_data_x1 <= pixel_data_temp_r;
			pixel_data_x <= pixel_data_r1(7 downto 0) & pixel_data_r2(7 downto 0);
		end if;
    end if;
	 
  end process get_pixel;




		------------------------------------------------------------------------------------------------------
		-- two-layer graphics subsystem
		------------------------------------------------------------------------------------------------------
	alpha1 <= alpha1b when pixel_data_r(15 downto 8)=opaquecolor else alpha1a;
   alpha2 <= "0000" when pixel_data_r(15 downto 8)=opaquecolor else alpha2a;
	

	
	 temp_in_b1<=(pixel_data_r(7 downto 6) & pixel_data_r(7)); --map blue to 3 bits resolution
	 temp_in_b2<=(pixel_data_r(15 downto 14) & pixel_data_r(15));
	 
	 -- multiply layers with brightness
	 mul_r1 : mul3 port map (x=>alpha1, y=>pixel_data_r(2 downto 0),p=>temp_r1);
	 mul_g1 : mul3 port map (x=>alpha1, y=>pixel_data_r(5 downto 3), p=>temp_g1);
	 mul_b1 : mul3 port map (x=>alpha1, y=>temp_in_b1, p=>temp_b1);
	 mul_r2 : mul3 port map (x=>alpha2, y=>pixel_data_r(10 downto 8), p=>temp_r2);
	 mul_g2 : mul3 port map (x=>alpha2, y=>pixel_data_r(13 downto 11), p=>temp_g2);
	 mul_b2 : mul3 port map (x=>alpha2, y=>temp_in_b2, p=>temp_b2);
	 
	 -- add layers
	 temp_r <= temp_r1+temp_r2;
	 temp_g <= temp_g1+temp_g2;
	 temp_b <= temp_b1+temp_b2;
	 
	 -- if saturated, then does not overflow
	 satur_r <= temp_r(3);
	 satur_r <= temp_g(3);
	 satur_r <= temp_b(3);
	 
	 pixel(2) <= (satur_r or temp_r(2)); 
	 pixel(1) <= (satur_r or temp_r(1)); 
	 pixel(0) <= (satur_r or temp_r(0)); 
	 pixel(5) <= (satur_g or temp_g(2));
	 pixel(4) <= (satur_g or temp_g(1));
	 pixel(3) <= (satur_g or temp_g(0));
	 pixel(8) <= (satur_b or temp_b(2));
	 pixel(7) <= (satur_b or temp_b(1));
    pixel(6) <= (satur_b or temp_b(0));



  -- map the current pixel to RGB values
  map_pixel : process(pixel, rgb_r, blank_r)
  begin
	
	-- 8-bit pixels map to RGB with reduced resolution in green component
   
	rgb_x <= pixel(2 downto 0) & pixel(5 downto 3) & pixel(8 downto 6);
	

    -- just blank the pixel if not in the visible region of the screen
    if blank_r(blank_r'high-1) = YES then
      rgb_x <= (others => '0');
    end if;

    -- break the pixel into its red, green and blue components
    r <= rgb_r(3*NUM_RGB_BITS-1 downto 2*NUM_RGB_BITS);
    g <= rgb_r(2*NUM_RGB_BITS-1 downto NUM_RGB_BITS);
    b <= rgb_r(NUM_RGB_BITS-1 downto 0);
  end process map_pixel;





-- update registers
  update : process(rst, clk)
  begin
    if rst = YES then
      eof_r          <= '0';
      rd_r           <= NO;
      hsync_r        <= (others => '1');
      blank_r        <= (others => '0');
		pixel_data_temp_r   <= (others => '0');
      pixel_data_r   <= (others => '0');
      pixel_data_r1   <= (others => '0');		
      pixel_data_r2   <= (others => '0');		
      rgb_r          <= (others => '0');
    elsif rising_edge(clk) then
      eof_r          <= eof_x;          -- end-of-frame signal goes at full clock rate to external system
      if cke = YES then
        rd_r         <= rd_x;
        hsync_r      <= hsync_x;
        blank_r      <= blank_x;
        pixel_data_r <= pixel_data_x;
		  pixel_data_temp_r <= pixel_data_temp_x;
        pixel_data_r1 <= pixel_data_x1;
		  pixel_data_r2 <= pixel_data_x2;
        rgb_r        <= rgb_x;
      end if;
    end if;
  end process update;

end architecture vga_arch;



---------------------------------------------------------
-- 2-LAYER SUBSYSTEM HELPER OBJECTS
---------------------------------------------------------
-- 3-bit multiplier
---------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
entity mul3 is
	port (
		x	:	in std_logic_vector(3 downto 0);
		y	:	in	std_logic_vector(2 downto 0);
		p  : out std_logic_vector(2 downto 0)
		);
end entity mul3;

architecture mul3_arch of mul3 is
	component half_adder is
		port (
			a		:	in std_logic;
			b		:	in	std_logic;
			res 	: out std_logic;
			cout 	: out std_logic
			);
	end component half_adder;

	component adder is
		port (
			a		:	in std_logic;
			b		:	in	std_logic;
			cin	:	in	std_logic;
			res 	: out std_logic;
			cout	: out	std_logic
			);
	end component adder;

	signal t2, t3, t4 : std_logic;
	signal c1, c21, c22, c31, c32, c41, c42: std_logic;
	signal s01, s02, s10, s11, s12, s20, s21, s22, s30, s31, s32 : std_logic;

begin


	s01 <= (x(0) and y(1));
	s02<= (x(0) and y(2));

	s10 <= (x(1) and y(0));
	s11<=(x(1) and y(1));
	s12 <= (x(1) and y(2));
	
	s20<= (x(2) and y(0));
	s21 <= (x(2) and y(1));
	s22 <= (x(2) and y(2));
	
	s30<= (x(3) and y(0));
	s31 <= (x(3) and y(1));
	s32 <= (x(3) and y(2));
	
	c1 <= s01 and s10;

	a1: half_adder port map (a=>s02, b=>s11, res=>t2, cout=>c21);

	c22 <= (t2 and c1) or (t2 and s20) or (s20 and c1);

	a2: adder port map (a=>s12, b=>s21, cin=>s30, res=>t3, cout=>c32);
	a3: adder port map (a=>c21, b=>c22, cin=>t3, res=>p(0), cout=>c31);

	a4: half_adder port map (a=>s22, b=>s31, res=>t4, cout=>c42);
	a5: adder port map (a=>c31, b=>c32, cin=>t4, res=>p(1), cout=>c41);

	p(2) <= c41 xor c42 xor s32;

end architecture mul3_arch;



---------------------------------------------------------
-- half adder
---------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
entity half_adder is
	port (
		a		:	in std_logic;
		b		:	in	std_logic;
		res 	: out std_logic;
		cout 	: out std_logic
		);
end entity half_adder;

architecture half_adder_arch of half_adder is
begin
	res<=a xor b;
	cout<=a and b;
end architecture half_adder_arch;




---------------------------------------------------------
-- full adder
---------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
entity adder is
	port (
		a		:	in std_logic;
		b		:	in	std_logic;
		cin	:	in	std_logic;
		res 	: out std_logic;
		cout	: out	std_logic
		);
end entity adder;

architecture adder_arch of adder is
begin
	res <= a xor b xor cin; 
	cout <= (a and b) or (b and cin) or (a and cin);
end architecture adder_arch;








---------------------------------------------------------
-- SYNC GENERATOR
---------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use IEEE.numeric_std.all;
use work.common.all;

-- Generate a sync pulse within a waveform PERIOD.
-- Also output the value of the counter used for timing so that
-- it can be used in generating an address for a video RAM.

entity sync is
  generic (
    FREQ    :     natural := 50_000;    -- master clock frequency (in KHz)
    PERIOD  :     natural := 32;        -- period of sync pulse (in us)
    START   :     natural := 26;        -- time sync pulse starts within the period (in us)
    WIDTH   :     natural := 4;         -- width of sync pulse (in us)
    VISIBLE :     natural := 1024       -- number of visible pixels/line or lines/frame
    );
  port (
    rst     : in  std_logic;            -- reset
    clk     : in  std_logic;            -- master clock
    cke     : in  std_logic;            -- clock-enable
    sync_n  : out std_logic;            -- sync pulse
    gate    : out std_logic;            -- single-clock pulse at start of sync pulse
    blank   : out std_logic;            -- blanking signal
    cnt     : out std_logic_vector(15 downto 0)  -- output the timing counter value
    );
end entity sync;


architecture sync_arch of sync is
  constant NORM             : natural := 1000;  -- normalization factor for us * KHz
  constant CYC_PERIOD       : natural := (PERIOD * FREQ)/NORM;  -- sync wave PERIOD in clock cycles
  constant CYC_START        : natural := (START * FREQ)/NORM;  -- sync pulse START in cycles
  constant CYC_WIDTH        : natural := (WIDTH * FREQ)/NORM;  -- sync pulse WIDTH in cycles
  constant CYC_END          : natural := CYC_START + CYC_WIDTH;  -- sync pulse end in cycles
  signal   cnt_r, cnt_x     : natural range 0 to (2**(cnt'high+1))-1;  -- counter for timing sync pulse waveform
--  signal   cnt_r, cnt_x     : std_logic_vector(cnt'range);  -- counter for timing sync pulse waveform
  signal   sync_r, sync_x   : std_logic;  -- sync register
  signal   gate_r, gate_x   : std_logic;  -- gate register
  signal   blank_r, blank_x : std_logic;  -- blank register
begin

-- increment counter and wrap around to zero at end of period
  cnt_x <= 0 when cnt_r = CYC_PERIOD-1 else cnt_r+1;

-- generate sync pulse within waveform period
  sync_x <= LO when cnt_r = CYC_START-1 else
            HI when cnt_r = CYC_END-1   else
            sync_r;
  sync_n <= sync_r;

-- generate gate signal at start of sync pulse
  gate_x <= YES when cnt_r = CYC_START-1 else NO;
  gate   <= gate_r;

-- generate blank signal after initial visible period
  blank_x <= YES when cnt_r = VISIBLE-1    else
             NO  when cnt_r = CYC_PERIOD-1 else
             blank_r;
  blank   <= blank_r;

-- output counter value
  cnt <= std_logic_vector(TO_UNSIGNED(cnt_r, cnt'length));

-- update counter and registers
  update : process(rst, clk)
  begin
    if rst = YES then
      cnt_r     <= 0;
      sync_r    <= HI;
      gate_r    <= NO;
      blank_r   <= YES;
    elsif rising_edge(clk) then
      if cke = YES then
        cnt_r   <= cnt_x;
        sync_r  <= sync_x;
        gate_r  <= gate_x;
        blank_r <= blank_x;
      end if;
    end if;
  end process update;

end architecture sync_arch;




---------------------------------------------------------
-- PIXEL PIPELINE FIFO
---------------------------------------------------------

library ieee, unisim;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use unisim.vcomponents.all;

entity fifo_cc is
  port (
    clk      : in  std_logic;
    rst      : in  std_logic;
    rd       : in  std_logic;
    wr       : in  std_logic;
    data_in  : in  std_logic_vector(15 downto 0);
    data_out : out std_logic_vector(15 downto 0);
    full     : out std_logic;
    empty    : out std_logic;
    level    : out std_logic_vector(7 downto 0)
    );
end entity fifo_cc;

architecture arch of fifo_cc is
  signal full_i   : std_logic;
  signal empty_i  : std_logic;
  signal rd_addr  : std_logic_vector(7 downto 0) := "00000000";
  signal wr_addr  : std_logic_vector(7 downto 0) := "00000000";
  signal level_i  : std_logic_vector(7 downto 0) := "00000000";
  signal rd_allow : std_logic;
  signal wr_allow : std_logic;
begin

  bram1 : RAMB4_S16_S16 port map (addra => rd_addr, addrb => wr_addr,
                                  dia   => (others => '0'), dib => data_in, wea => '0', web => '1',
                                  clka  => clk, clkb => clk, rsta => '0', rstb => '0',
                                  ena   => rd_allow, enb => wr_allow, doa => data_out );

  rd_allow <= rd and not empty_i;
  wr_allow <= wr and not full_i;

  process (clk, rst)
  begin
    if rst = '1' then
      rd_addr   <= (others => '0');
      wr_addr   <= (others => '0');
      level_i   <= (others => '0');
    elsif rising_edge(clk) then
      if rd_allow = '1' then
        rd_addr <= rd_addr + '1';
      end if;
      if wr_allow = '1' then
        wr_addr <= wr_addr + '1';
      end if;
      if (wr_allow and not rd_allow and not full_i) = '1' then
        level_i <= level_i + '1';
      elsif (rd_allow and not wr_allow and not empty_i) = '1' then
        level_i <= level_i - '1';
      end if;
    end if;
  end process;

  full_i  <= '1' when level_i = "11111111" else '0';
  full    <= full_i;
  empty_i <= '1' when level_i = "00000000" else '0';
  empty   <= empty_i;
  level   <= level_i;

end architecture arch;
